/*{{{  includes and defines*/
#include <stdio.h>

#define BOOT_FILE_VERSION_SUPPORTED         1
#define BOOT_FILE_VERSION_1                 1
#define BOOT_FILE_VERSION_2                 2


#define VERSION_RECORD_SIZE                 8 /* Tag and version*/

#define VERSION_RECORD_TAG                  16
#define NETWORK_LOADER_RECORD_TAG           17
#define CFB_FILENAME_RECORD_TAG             18

#define HOST_INFORMATION_RECORD_TAG         19
#define ROUTER_HEADER_RECORD_TAG            20

#define H1_COMMENT_RECORD_TAG_SIZE          4

#define MY_UNDEFINED_VALUE                  0xFFFFFFFF
#define MOST_NEG_INT                        0x80000000

#define MAX_HEADER_LEN			    32

#define MAX_HEADER_STRLEN		    (3 * MAX_HEADER_LEN)


#define BOOT_HEADER_IN			"0"
#define BOOT_HEADER_OUT			"0"
#define BOOT_HEADER_SIZE		1
/*}}}*/

/*{{{  globals*/

int NumberOfRouters, OutHeaderLen, InHeaderLen, VirtualOutHeaderLen, VirtualInHeaderLen, BootRoutingHeaderLen;

char RoutingHeader[MAX_HEADER_LEN];
char BootRoutingHeader[MAX_HEADER_LEN];
char VLCB_HeaderOut[MAX_HEADER_LEN];
char VLCB_HeaderIn[MAX_HEADER_LEN];

char BootDeviceName[256];
char BTL_DeviceName[256];
extern char DeviceName[256];
/*}}}*/

/*{{{  macros*/
#define SKIP_RECORD( fp, Len ) fseek( fp, Len, SEEK_CUR )


#define SET_DEVICE_NAME( DevName, InLen, OutRoutLen, OutLen, InHead, OutRoutHead, OutHead ) {\
	char LinkStr[4], tempStr[4], OutHeader[MAX_HEADER_STRLEN];	\
	char InHeader[3];						\
	int i;								\
	LinkStr[0] = DeviceName[0];					\
	LinkStr[1] = DeviceName[1];					\
	LinkStr[2] = '\0';						\
	OutHeader[0] = '\0';						\
	for ( i = 0; i < OutRoutLen; i++ ){				\
		sprintf( tempStr, "%x.", OutRoutHead[i] );		\
		strcat( OutHeader, tempStr );				\
	}								\
	sprintf( tempStr, "%x", OutHead );				\
	strcat( OutHeader, tempStr );					\
	sprintf( InHeader, "%x", InHead );				\
	sprintf( DevName, "%s.%s %s %s.%s %s", LinkStr, OutHeader, InHeader, LinkStr, InHeader, OutHeader );\
}

/*}}}*/

/*{{{  ProcessComment*/

int ProcessComment( FILE *fp )
{
	int TotalLen = 0;
	int Length;
	unsigned long Tag;
	unsigned long Version;
	int i;
	char BootLinkName[4], tempStr[4], BootOutHeader[MAX_HEADER_STRLEN];

	while (1) {
		if ( fread( &Length, sizeof(int), 1, fp ) != 1 ) return(-1);
		TotalLen += Length;
		if ( fread( &Tag, sizeof(int), 1, fp ) != 1 ) return(-1);
		switch ( Tag ){
		case VERSION_RECORD_TAG: 
			if ( fread( &Version, sizeof(int), 1, fp ) != 1 )
				return(-1);
			SKIP_RECORD( fp, Length-8 );break;
		case HOST_INFORMATION_RECORD_TAG :
			if ( fread( &NumberOfRouters, sizeof(int), 1, fp ) != 1 )
				return(-1);
			break;
		case ROUTER_HEADER_RECORD_TAG :
			Length -= 4;
			for ( i = 0; i < NumberOfRouters; i++ ){
				if ( fread( &OutHeaderLen, sizeof(int), 1, fp ) != 1 )
					return(-1);
				Length = Length - (4+OutHeaderLen);
				if ( OutHeaderLen > 0 )
					if ( fread( RoutingHeader, sizeof(char), OutHeaderLen, fp ) != OutHeaderLen )
						return(-1);
				if ( fread( &InHeaderLen, sizeof(int), 1, fp ) != 1 )
					return(-1);
				if ( InHeaderLen > 0 )
					fseek( fp, InHeaderLen, SEEK_CUR );
				Length = Length - (4+InHeaderLen);
				if ( fread( &VirtualOutHeaderLen, sizeof(int), 1, fp ) != 1 )
					return(-1);
				if ( VirtualOutHeaderLen > 0 )
					if ( fread( VLCB_HeaderOut, sizeof(char), VirtualOutHeaderLen, fp ) != VirtualOutHeaderLen )
						return(-1);
				Length = Length - (4+VirtualOutHeaderLen);
				if ( fread( &VirtualInHeaderLen, sizeof(int), 1, fp ) != 1 )
					return(-1);
				if ( VirtualInHeaderLen > 0 )
					if ( fread( VLCB_HeaderIn, sizeof(char), VirtualInHeaderLen, fp ) != VirtualInHeaderLen )
						return(-1);
				Length = Length - (4+VirtualInHeaderLen);
				if ( i == 0 ){
					BootRoutingHeaderLen = OutHeaderLen;
					memcpy( BootRoutingHeader, RoutingHeader, OutHeaderLen );

				}
/*
				if ((Version ==1)||( i > 0)){
					SET_DEVICE_NAME( BTL_DeviceName,
						VirtualInHeaderLen,
						VirtualOutHeaderLen,OutHeaderLen,
						VLCB_HeaderIn,
						RoutingHeader,
						VLCB_HeaderOut );

				}
*/
				SKIP_RECORD( fp, Length );
				
			}
/*
			SET_DEVICE_NAME( BootDeviceName,
				1, BootRoutingHeaderLen,1, BOOT_HEADER_IN,
				BootRoutingHeader, BOOT_HEADER_OUT );
*/
			BootLinkName[0] = DeviceName[0];
			BootLinkName[1] = DeviceName[1];
			BootLinkName[2] = '\0';
			BootOutHeader[0] = '\0';
			for ( i = 0; i < BootRoutingHeaderLen; i++ ){
				sprintf( tempStr, "%x.", BootRoutingHeader[i] );
				strcat( BootOutHeader, tempStr );
			}
			strcat( BootOutHeader, BOOT_HEADER_OUT );
			sprintf( BootDeviceName, "%s.%s %s %s.%s %s", BootLinkName, BootOutHeader, BOOT_HEADER_IN, BootLinkName, BOOT_HEADER_IN, BootOutHeader );
			return(0);
		default: SKIP_RECORD( fp, Length-4 );
		}
	}
}
/*}}}*/

