
#ifndef VCP_QUEUE_H
#define VCP_QUEUE_H

/* Generic Queue manipulation routines 
   NOTE: All elements of Queue should have QUEUE_CHAIN
   as the first element
*/

 
#include "systype.h"

/*{{{  QEmpty*/
/*	Is the queue empty ? 
*/
#define QEmpty( q )		( ((void *)q) == NULL )
/*}}}*/

/*{{{  InitQ*/
/*	Initialize a queue
*/
#define	InitQ( q )		q = NULL
/*}}}*/

/*{{{  EnQueue*/
/* 	Add 'element' to the tail of the specified queue
*/
#define EnQueue( q, element )	{					\
	((QUEUE_CHAIN *)element)->Next = NULL;				\
	if ( QEmpty(q) ) q = element;					\
	else {								\
		QUEUE_CHAIN *temp;					\
		temp = (QUEUE_CHAIN *)q;				\
		while ( temp->Next != NULL )				\
			temp = temp->Next;				\
		temp->Next = (QUEUE_CHAIN *)element;			\
	}								\
}
/*}}}*/

/*{{{  DeQueue*/
/* 	Obtain an element from the head of the specified queue
*/
#define DeQueue( q, element )	{					\
	if ( QEmpty(q) ) element = NULL;				\
	else {								\
		element = q;						\
		q = ((QUEUE_CHAIN *)q)->Next;				\
	}								\
}
/*}}}*/

/*{{{  QHead*/
/* 	Return the element at the head of the specified queue
*/
#define QHead( q ) ( q )

/*}}}*/

/*{{{  QNext*/
/*	Return the element after the specified element in the queue
*/
#define	QNext( element ) ((element == NULL)?NULL:			\
				((QUEUE_CHAIN *)element)->Next)
/*}}}*/

/*{{{  QRemove*/
/*	Remove the specified element from the queue. Do not free the element.
*/
#define QRemove( q, element ) {						\
	QUEUE_CHAIN *temp = q;						\
	if ( q == element )						\
		q = ((QUEUE_CHAIN *)element)->Next;			\
	else {								\
		while (( temp != NULL) && ( temp->Next!= (void *)element ))	\
			temp = temp->Next;				\
		if ( temp != NULL )					\
			temp->Next = temp->Next->Next;			\
	}								\
}
/*}}}*/

/*{{{  QDelete*/
/*	Delete the specified element from the queue and free the memory
 	allocated to it.
*/
#define QDelete( q, element ) {						\
	QRemove( q, element );						\
	vcp_free( element );						\
}
/*}}}*/

/*{{{  InsertQ*/
/*	Add an element to the Queue after 'prev'
*/
#define InsertQ( q, element, prev ){					\
	if ( q == NULL ) { q = element; 				\
		((QUEUE_CHAIN *)element)->Next = NULL; }		\
	else if ( prev == NULL ) { 					\
		((QUEUE_CHAIN *)element)->Next = q; 			\
		q = element; 						\
	}								\
	else {								\
		((QUEUE_CHAIN *)element)->Next = ((QUEUE_CHAIN *)prev)->Next;\
		((QUEUE_CHAIN *)prev)->Next = element;			\
	}								\
}
/*}}}*/

/*	Table manipulation routines
 *	NOTE : All table elements should have QUEUE_CHAIN as the first element
 */

/*{{{  InitTable*/
/*	Initialize a table
*/
#define InitTable( table ) {						\
	int i;								\
	for ( i = 0; i < MAX_TABLE_SIZE; i++ ) 				\
		table[i] = NULL;					\
}
/*}}}*/

/*{{{  AddTable*/
/*	Add an element with the specified key to the table
*/
#define AddTable( table, key, element ) 			\
	EnQueue( table[key % MAX_TABLE_SIZE], element );
/*}}}*/

/*{{{  DeleteTable*/
/*	Delete the element with the specified key from the table
*/
#define DeleteTable( table, key, element )			\
	QDelete( table[key % MAX_TABLE_SIZE], element );
/*}}}*/

/*{{{  RemoveTable*/
/*	Remove the element with the specified key from the table
 	Do not free the element
*/
#define RemoveTable( table, key, element )			\
	QRemove( table[key % MAX_TABLE_SIZE], element );
/*}}}*/
	
#endif
