
/*{{{  include files*/

#include <stdio.h>
#include <conc.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <netdb.h>
#include <errno.h>
#include "vcpdefs.h"
#include "queue.h"
#include "tclVcpDefs.h"
#include "tclVcpType.h"
#include "tclVcpMac.h"
#include "tclVcpExt.h"
#include "tclInt.h"
#include "debug.h"
/*}}}*/

/*{{{  Commands bound to VCP events */
int NumTclPending = 0;
extern Semaphore TclSemaphore;
WDESC_TYPE TclProcDesc=NOT_PROCESS;
int IServerSocket = -1;
#define QLEN		10
void *vcp_malloc(int size );
void vcp_free( void *ptr );
void *vcp_realloc( void *ptr, int size );
/*
 *------------------------------------------------------------------------
 *
 *	TclPending:
 *		Return number of commands bound to VCP events that are
 *		still pending
 *
 *------------------------------------------------------------------------
 */
int TclPending()
{
	return( NumTclPending );
}
/*}}}*/

/*{{{  ConnectServerToWindow*/
/*
 *------------------------------------------------------------------------
 *
 *	ConnectServerToWindow:
 *		Open a window using the specified Unix shell command,
 *	Create a socket and connect it to the window.
 *
 *	Return value :
 *		Socket Id if successful, negative value otherwise.
 *
 *------------------------------------------------------------------------
 */
#undef ntohs
int ConnectServerToWindow( Tcl_Interp *interp, char *WinCmd )
{
	struct sockaddr_in name;
	int sock, msock;
	int length;
	char *CommandBuf;
	int i, j;

	sock = socket( AF_INET, SOCK_STREAM, 0 );
	if ( sock < 0 ){
		fprintf( stderr, "Cannot create socket for iserver, err %d\n", errno );
		return( -1 );
	}
	bzero((char *)&name, sizeof(name));
	name.sin_family = AF_INET;
	name.sin_addr.s_addr = INADDR_ANY;
	name.sin_port = 0;
	if ( bind( sock, (struct sockaddr *)&name, sizeof(name)) < 0 ){
		fprintf( stderr, "Cannot bind socket to port for iserver, err %d\n", errno );
		return( -1 );
	}
        length = sizeof(name);
        if ( getsockname( sock, (struct sockaddr *)&name, &length) < 0 ){
                fprintf( stderr, "error getting socket name\n" );
                return(-1);
        }

	if ( listen(sock , QLEN ) < 0 ){
		fprintf( stderr, "Cannot listen to iserver port, err %d\n", errno );
		return( -1 );
	}
	if ( WinCmd[0] != '\0' ) {
		CommandBuf = (char *)vcp_malloc(MAX_UNIX_COMMAND_LEN * sizeof(char));
		if ( CommandBuf == NULL ){
			fprintf( stderr, "Out of memory, cannot execute window command\n" );
			return(-1);
		}
		i = 0;j = 0;
		while ( WinCmd[i] != '\0' ){
			if ( WinCmd[i] == PARAMETER_SPEC_CHAR ){
				if ( strncmp( &(WinCmd[i+1]), HOST_NAME , strlen(HOST_NAME)) == 0 ){
					gethostname( &(CommandBuf[j]), MAX_UNIX_COMMAND_LEN-j-1 );
					i  += (strlen(HOST_NAME)+1);
					while ( CommandBuf[j] != '\0' )j++;
				}
				if ( strncmp( &(WinCmd[i+1]), PORT_NUMBER , strlen(PORT_NUMBER)) == 0 ){
					i  += (strlen(PORT_NUMBER)+1);
					CommandBuf[j] = '\0';
					sprintf( CommandBuf, "%s %d ", CommandBuf, ntohs(name.sin_port));
					while ( CommandBuf[j] != '\0' )j++;
				}
				else 
					CommandBuf[j++] = WinCmd[i++];
			}
			else 
				CommandBuf[j++] = WinCmd[i++];
		}
		CommandBuf[j] = '\0';
		sprintf( CommandBuf, "%s &\n", CommandBuf );
		system( CommandBuf );
	}
        length = sizeof(name);
        msock = accept( sock, (struct sockaddr *)&name, &length );
        if ( msock < 0 )
                fprintf( stderr, "Could not connect iserver to window\n" );
        return ( msock );
}
/*}}}*/


/*{{{  Start iserver */
/*
 *------------------------------------------------------------------------
 *
 *	StartServer:
 *		Start iserver with the given parameters. If the -window
 *	option is specified, the Unix command <WinCmd> is executed, with
 *	hostname and a portid as parameters. StdIn, StdOut and StdErr are
 *	connected to that portid. 
 *	If -bind option is specified, the server returns after booting the
 *	program, after creating a server thread to serve it.
 *	Command Format:
 *		run #servinit <link spec> [-window <WinCmd>]
 *			[-bind <BindCmd>] <parameters>
 *	All the parameters following the run command are passed to the
 *	server.
 *
 *	Returns nothing
 *
 *------------------------------------------------------------------------
 */
int StartServer( Tcl_Interp *interp, int argc, char **argv )
{
	char *LinkName;
	int i, j;
	int NewStdOut=-1, NewStdIn=-1, NewStdErr=-1;
	int ArgIndex;
	char *BoundCommand = NULL;
	int ServerId;
	char *ServerName = argv[1];

	LinkName = argv[2];
	if ( strlen(LinkName) == 2 ){
		/* iserver waiting on raw link */
		CHECK_LINK_NUMBER( argv[2], interp );
		ArgIndex = 3;
	}
	else {
		/* iserver waiting on a virtual channel */
		/* Parse virtual channel info */
		CHECK_HEADER( argv[2], False, interp );
		CHECK_VCHAN( argv[3], True, interp );
		CHECK_VCHAN( argv[4], False, interp );
		CHECK_HEADER( argv[5], True, interp );
		LinkName = (char *)vcp_malloc(( strlen(argv[2])+strlen(argv[3])+strlen(argv[4])+strlen(argv[5])+8)*sizeof(char));
		sprintf( LinkName, "%s %s %s %s", argv[2], argv[3], argv[4], argv[5] );
		ArgIndex = 6;
	}
	if ( strcmp( argv[ArgIndex], WINDOW_OPTION ) == 0 ){
		NewStdIn = NewStdOut = NewStdErr = 
			ConnectServerToWindow( interp, argv[ArgIndex+1] );
		ArgIndex += 2;
	}
	if ( strcmp( argv[ArgIndex], BIND_OPTION ) == 0 ){
		BoundCommand = (char *)vcp_malloc(strlen(argv[ArgIndex+1])+1);
		if ( BoundCommand == NULL )
			fprintf( stderr, "Cannot run server in unblocked mode, out of memory\n" );
		strcpy( BoundCommand, argv[ArgIndex+1] );
		SemWait( &TclSemaphore );
		NumTclPending++;
		SemSignal( &TclSemaphore );
		ArgIndex += 2;
	}
	/* Pass all arguments except "run" and 
	   virtual channel info to server */
	argv[0] = argv[1];
	for ( i = ArgIndex, j = 1; i < argc; i++ ){
		if (( argv[i] != NULL )&&( argv[i][0] != '\0' ))
			argv[j++] = argv[i];
	}
	argc = j;
	ServerId = ServerInit( ServerName, NewStdIn, NewStdOut, NewStdErr, BoundCommand, LinkName, argc, argv );
	return( ServerId );
}
/*}}}*/

/*{{{  TCL Run command*/
/*
 *------------------------------------------------------------------------
 *
 *	Tcl_RunCmd :
 *		Execute the given function with the specified parameters
 *
 *	Command format is : run #<function> <parameters>
 *
 *	Return value :
 *		TCL_OK if successful, TCL_ERROR otherwise
 *
 *------------------------------------------------------------------------
 */
int
Tcl_RunCmd(
    ClientData dummy,			/* Not used. */
    register Tcl_Interp *interp,	/* Current interpreter. */
    int argc,				/* Number of arguments. */
    char **argv)			/* Argument strings. */
{
	int LinkNumber;
	int ServerId;
	char ServName[10];

	if (( strcmp( argv[1], SERV_INIT ) == 0 )||
	    ( strcmp( argv[1], COFF_SERV_INIT ) == 0 )){
		ServerId = StartServer( interp, argc, argv );
		sprintf( ServName, "%d", ServerId );
		Tcl_AppendResult( interp,ServName , NULL );
		return TCL_OK;
	}
	else if ( strcmp( argv[1], LOCAL_WRITE ) == 0 ){
		unsigned long Address, Data;
		if ( argc != 4 ){
			Tcl_AppendResult( interp, "wrong # args: run #localwrite <address> <data>\n", NULL );
			return TCL_ERROR;
			
		}
		CHECK_HEXVAL( argv[2], interp );
		CHECK_HEXVAL( argv[3], interp );
		GET_HEXVAL( argv[2], Address );
		GET_HEXVAL( argv[3], Data );
		if ( strlen( argv[3]) <= 2 ){
			*(unsigned char *)Address = Data;
		}
		if ( strlen( argv[3]) <= 4 )
			*(unsigned short *)Address = Data;
		else
			*(unsigned long *)Address = Data;
		return TCL_OK;
		
	}
	if ( argc != 3 ) {
		Tcl_AppendResult( interp, "wrong # args: run <function> <link>\n", NULL );
		return TCL_ERROR;
	}
	if ( strcmp( argv[1], VCP_INIT ) == 0 ){
		CHECK_LINK_NUMBER( argv[2], interp );
		GET_LINK_NUMBER( argv[2], LinkNumber );
		vcpinit( LinkNumber );
	}
	else if ( strcmp( argv[1], VCP_QUIT ) == 0 ){
		CHECK_LINK_NUMBER( argv[2], interp );
		GET_LINK_NUMBER( argv[2], LinkNumber );
		vcpquit( LinkNumber );
	}
	else if ( strcmp( argv[1], SERV_QUIT ) == 0 ){
		ServerQuit( atoi(argv[2]) );
	}
	else {
		Tcl_AppendResult( interp, "ERROR : Unknown function\n", NULL );
		return TCL_ERROR;
	}
	return TCL_OK;
}
/*}}}*/

/*{{{  TCL Send command*/
/*
 *------------------------------------------------------------------------
 *
 *	Tcl_SendCmd :
 *		Send some data or file to a specified virtual channel
 *	destination.  
 *
 *	Command format is : vcpsend <header> <vc> [-data <data>| -file <filename>]
 *				[-bind <command>] [-timeout <time> <command>]
 *	If -bind option is specified, the call does not block, otherwise the
 *	call returns only after the complete message has been sent and the
 *	acknowledgement received.. The TCL command specified with the -bind 
 *	option is executed after the complete message is sent. 
 *	If -timeout option is specified, wait only for <time> seconds to send
 *	the message. If unsuccessful, execute the specified timeout command.
 *
 *	Return value :
 *		TCL_OK if successful, TCL_ERROR otherwise
 *
 *------------------------------------------------------------------------
 */
int
Tcl_SendCmd(
    ClientData dummy,			/* Not used. */
    register Tcl_Interp *interp,	/* Current interpreter. */
    int argc,				/* Number of arguments. */
    char **argv)			/* Argument strings. */
{
	Boolean VCP_Send, UnBlockedSend;
	int ArgIndex;
	FILE_OR_BUFFER FileOrBuf;
	int Link1, Link2;
	int LinkNumber;
	int TimeOut = -1;
	char *TimeOutCommand, *BindCommand;
	int result;

	/*{{{  Parse command*/
	if( strlen( argv[1] ) > 2 ){
		if ( argc < 5 ) SEND_COMMAND_ERROR(interp);
		CHECK_HEADER( argv[1], False, interp );
		CHECK_VCHAN( argv[2], True, interp );
		VCP_Send = True;
		Link1 = -1;
		GET_LINK_NUMBER( argv[2], Link1 );
		if ( Link1 != -1 ){
			GET_LINK_NUMBER( argv[1], Link2 );
			if ( Link2 != Link1 ){
				Tcl_AppendResult( interp, "ERROR : Send and receive link numbers are different\n", NULL );
				return TCL_ERROR;
			}
		}
		ArgIndex = 3;
	}
	else {
		if ( argc < 4 ) SEND_COMMAND_ERROR(interp);
		CHECK_LINK_NUMBER( argv[1], interp );
		GET_LINK_NUMBER( argv[1], LinkNumber );
		VCP_Send = False;
		ArgIndex = 2;
	}
	if ( argc < ArgIndex+2 ) SEND_COMMAND_ERROR(interp);
	if ( strcmp( argv[ArgIndex], DATA_OPTION ) == 0 ){
		CHECK_BINARY_DATA( argv[ArgIndex+1], interp );
		FileOrBuf.FileFlag = False;
		FileOrBuf.BufferPtr = argv[ArgIndex+1];
	}
	else if ( strcmp( argv[ArgIndex], FILE_OPTION ) == 0 ){
		FileOrBuf.FileFlag = True;
		FileOrBuf.BufferPtr = argv[ArgIndex+1];
	}
	else SEND_COMMAND_ERROR( interp );
	ArgIndex += 2;
	/*}}}*/
	UnBlockedSend = False; 
	BindCommand = NULL; TimeOutCommand = NULL;
	TimeOut = -1;
	if ( argc > ArgIndex ) {
		UnBlockedSend = ( strcmp( argv[ArgIndex], BIND_OPTION ) == 0 );
		if ( UnBlockedSend ){
			BindCommand = argv[ArgIndex+1];
			ArgIndex += 2;
		}
	}
	if ( argc > ArgIndex ) {
		if (( strcmp( argv[ArgIndex], TIMEOUT_OPTION ) != 0 )
		    || ( argc < ArgIndex+3))
			SEND_COMMAND_ERROR(interp);
		TimeOut = atoi( argv[ArgIndex+1] );
		TimeOutCommand = (char *)vcp_malloc(strlen(argv[ArgIndex+2])+1);
		strcpy(TimeOutCommand, argv[ArgIndex+2]);
	}
	if ( UnBlockedSend ){
		NumTclPending++;
		if ( VCP_Send ){
			if ( VCP_UnBlockedSend( argv[1], argv[2], &FileOrBuf, BindCommand, TimeOut, TimeOutCommand ) < 0 ){
				Tcl_AppendResult( interp, "Out of memory\n", NULL );
				return TCL_ERROR;
			}
		}
		else
			result = Chan_UnBlockedSend( LinkNumber, &FileOrBuf, BindCommand, TimeOut, TimeOutCommand );
	}
	else {
		if ( VCP_Send )
			result = VCP_BlockedSend( argv[1], argv[2], &FileOrBuf, TimeOut, TimeOutCommand );
		else 
			result = Chan_BlockedSend( LinkNumber, &FileOrBuf, TimeOut, TimeOutCommand );
	}
	if ( result == VCP_TimeOut )
		Tcl_AppendResult( interp, "Timeout", NULL );
	if ( result < 0 )return( TCL_ERROR );
	return( TCL_OK );
}

/*}}}*/

/*{{{  TCL Recv command*/
/*
 *------------------------------------------------------------------------
 *
 *	Tcl_RecvCmd :
 *		Receive a message onto a file or a TCL variable.
 *
 *	Command format is : recv <vc> <header> [-data <var>| -file <filename>]
 *				[-bind <command>] [-timeout <time> <command>]
 *	If -bind option is specified, the call does not block, otherwise the
 *	call returns only after the complete message is received. The TCL
 *	command specified with the -bind option is executed after the 
 *	complete message is received.
 *	If -timeout option is specified, wait only for <time> seconds to receive
 *	the message. If unsuccessful, execute the specified timeout command.
 *
 *	Return value :
 *		TCL_OK if successful, TCL_ERROR otherwise
 *
 *------------------------------------------------------------------------
 */
int
Tcl_RecvCmd(
    ClientData dummy,			/* Not used. */
    register Tcl_Interp *interp,	/* Current interpreter. */
    int argc,				/* Number of arguments. */
    char **argv)			/* Argument strings. */
{
	Boolean VCP_Recv, UnBlockedRecv;
	int ArgIndex;
	FILE_OR_BUFFER FileOrBuf;
	char *RecvVar = NULL;
	char *tempStr;
	int Link1, Link2;
	int LinkNumber;
	int count, ret;
	int TimeOut = -1;
	char *TimeOutCommand, *BindCommand;

	/*{{{  Parse command*/
	if( strlen( argv[1] ) > 2 ){
		CHECK_VCHAN( argv[1], False, interp );
		CHECK_HEADER( argv[2], True, interp );
		VCP_Recv = True;
		Link1 = -1;
		GET_LINK_NUMBER( argv[2], Link1 );
		if ( Link1 != -1 ){
			GET_LINK_NUMBER( argv[1], Link2 );
			if ( Link2 != Link1 ) {
				Tcl_AppendResult( interp, "ERROR : Send and receive link numbers are different\n", NULL );
				return TCL_ERROR;
			}
		}
		ArgIndex = 3;
		VCP_Recv = True;
	}
	else {
		CHECK_LINK_NUMBER( argv[1], interp );
		GET_LINK_NUMBER( argv[1], LinkNumber );
		sscanf( argv[2], "%d", &count );
		ArgIndex = 3;
		VCP_Recv = False;
	}
	if ( strcmp( argv[ArgIndex], DATA_OPTION ) == 0 ){
		FileOrBuf.FileFlag = False;
		FileOrBuf.BufferPtr = NULL;
		RecvVar = argv[ArgIndex+1];
	}
	else if ( strcmp( argv[ArgIndex], FILE_OPTION ) == 0 ){
		FileOrBuf.FileFlag = True;
		FileOrBuf.BufferPtr = argv[ArgIndex+1];
	}
	else RECV_COMMAND_ERROR( interp );
	ArgIndex += 2;
	/*}}}*/
	UnBlockedRecv = False; 
	BindCommand = NULL; TimeOutCommand = NULL;
	TimeOut = -1;
	if ( argc > ArgIndex ) {
		UnBlockedRecv = ( strcmp( argv[ArgIndex], BIND_OPTION ) == 0 );
		if ( UnBlockedRecv ){
			BindCommand = argv[ArgIndex+1];
			ArgIndex += 2;
		}
	}
	if ( argc > ArgIndex ) {
		if (( strcmp( argv[ArgIndex], TIMEOUT_OPTION ) != 0 )
		    || ( argc < ArgIndex+3))
			RECV_COMMAND_ERROR(interp);
		TimeOut = atoi( argv[ArgIndex+1] );
		TimeOutCommand = (char *)vcp_malloc(strlen(argv[ArgIndex+2])+1);
		strcpy(TimeOutCommand, argv[ArgIndex+2]);
	}
	if ( UnBlockedRecv ) {
		SemWait( &TclSemaphore );
		if ( FileOrBuf.FileFlag )
			NumTclPending++;
		else
			NumTclPending += 2;
		SemSignal( &TclSemaphore );
		if ( VCP_Recv ){
			if ((ret = VCP_UnBlockedReceive( argv[1], argv[2], &FileOrBuf, BindCommand, RecvVar, TimeOut, TimeOutCommand )) < 0 ){
				Tcl_AppendResult( interp, "Out of memory\n", NULL );
				return TCL_ERROR;
			}
		}
		else
			ret = Chan_UnBlockedReceive( LinkNumber, &FileOrBuf, count, BindCommand, RecvVar, TimeOut, TimeOutCommand );
	}
	else {
		if ( VCP_Recv ) 
			ret = VCP_BlockedReceive( argv[1], argv[2], &FileOrBuf, TimeOut, TimeOutCommand );
		else
			ret = Chan_BlockedReceive( LinkNumber, &FileOrBuf, count, TimeOut, TimeOutCommand );
		if (( FileOrBuf.FileFlag == False )&&( ret == VCP_OK )) {
			char *result;
			result = Tcl_SetVar(interp, RecvVar, FileOrBuf.BufferPtr, TCL_LEAVE_ERR_MSG);
			vcp_free( FileOrBuf.BufferPtr );
			if (result == NULL) {
			    return TCL_ERROR;
			}
			interp->result = result;
		}
	}

	if ( ret == VCP_TimeOut )
		Tcl_AppendResult( interp, "Timeout", NULL );
	if ( ret < 0 ) return(TCL_ERROR);
	return( TCL_OK );
}

/*}}}*/


/*{{{  TCL Transfer command*/
/*
 *------------------------------------------------------------------------
 *
 *	Tcl_XferCmd :
 *		Send some data or file to a specified virtual channel
 *	destination.  
 *
 *	Command format is : vcpsend <header> <vc> [-data <data>| -file <filename>]
 *				[-bind <command>] [-timeout <time> <command>]
 *	If -bind option is specified, the call does not block, otherwise the
 *	call returns only after the complete message has been sent and the
 *	acknowledgement received.. The TCL command specified with the -bind 
 *	option is executed after the complete message is sent. 
 *	If -timeout option is specified, wait only for <time> seconds to send
 *	the message. If unsuccessful, execute the specified timeout command.
 *
 *	Return value :
 *		TCL_OK if successful, TCL_ERROR otherwise
 *
 *------------------------------------------------------------------------
 */
int
Tcl_XferCmd(
    ClientData dummy,			/* Not used. */
    register Tcl_Interp *interp,	/* Current interpreter. */
    int argc,				/* Number of arguments. */
    char **argv)			/* Argument strings. */
{
	Boolean VCP_Xfer;
	int ArgIndex;
	char *Buffer;
	int Link1, Link2;
	int LinkNumber;
	int TimeOut = -1;
	char *TimeOutCommand;
	int result;

	/*{{{  Parse command*/
	if( strlen( argv[1] ) > 2 ){
		if ( argc < 7 ) XFER_COMMAND_ERROR(interp);
		CHECK_HEADER( argv[1], False, interp );
		CHECK_VCHAN( argv[2], True, interp );
		CHECK_VCHAN( argv[3], True, interp );
		CHECK_HEADER( argv[4], True, interp );
		VCP_Xfer = True;
		Link1 = -1;
		GET_LINK_NUMBER( argv[2], Link1 );
		if ( Link1 != -1 ){
			GET_LINK_NUMBER( argv[1], Link2 );
			if ( Link2 != Link1 ){
				Tcl_AppendResult( interp, "ERROR : Send and receive link numbers are different\n", NULL );
				return TCL_ERROR;
			}
		}
		ArgIndex = 5;
	}
	else {
		XFER_COMMAND_ERROR(interp);
	}
	if ( argc < ArgIndex+2 ) XFER_COMMAND_ERROR(interp);
	if ( strcmp( argv[ArgIndex], DATA_OPTION ) == 0 ){
		Buffer = argv[ArgIndex+1];
	}
	ArgIndex += 2;
	/*}}}*/
	TimeOutCommand = NULL;
	TimeOut = -1;
	if ( argc > ArgIndex ) {
		if (( strcmp( argv[ArgIndex], TIMEOUT_OPTION ) != 0 )
		    || ( argc < ArgIndex+3))
			XFER_COMMAND_ERROR(interp);
		TimeOut = atoi( argv[ArgIndex+1] );
		TimeOutCommand = (char *)vcp_malloc(strlen(argv[ArgIndex+2])+1);
		strcpy(TimeOutCommand, argv[ArgIndex+2]);
	}
	if ( VCP_Xfer )
		result = VCP_BlockedXfer( argv[1], argv[2], argv[3], argv[4], Buffer, TimeOut, TimeOutCommand );
	if ( result == VCP_TimeOut )
		Tcl_AppendResult( interp, "Timeout", NULL );
	if ( result < 0 )return( TCL_ERROR );
	return( TCL_OK );
}

/*}}}*/

/*{{{  TCL Inspect command*/
/*
 *------------------------------------------------------------------------
 *
 *	Tcl_InspectCmd :
 *		Return the virtual channel headers for which messages have
 *		arrived.
 *
 *	Command format is : inspect <link> <variable>
 *
 *	Return value :
 *		TCL_OK if successful, TCL_ERROR otherwise
 *
 *------------------------------------------------------------------------
 */
int
Tcl_InspectCmd(
    ClientData dummy,			/* Not used. */
    register Tcl_Interp *interp,	/* Current interpreter. */
    int argc,				/* Number of arguments. */
    char **argv)			/* Argument strings. */
{
	int LinkNumber;
	char *result;
	char *HeaderList;

	if ( argc != 3 ) return TCL_ERROR;
	CHECK_LINK_NUMBER( argv[1] , interp);
	GET_LINK_NUMBER( argv[1], LinkNumber );
	HeaderList = InspectReceiveQ( LinkNumber );
	result = Tcl_SetVar(interp, argv[2], HeaderList, TCL_LEAVE_ERR_MSG);
	vcp_free( HeaderList );
	if (result == NULL) {
	    return TCL_ERROR;
	}
	interp->result = result;
	
	return TCL_OK;
}
/*}}}*/

/*{{{  Execute a TCL command*/
/*
 *------------------------------------------------------------------------
 *
 *	ExecuteTclCommand :
 *		Executes TCL commands from TclCommandList.
 *		Also signals the ServerSemaphore, so that the server can
 *	execute, if it is ready.
 *
 *	Returns nothing
 *
 *------------------------------------------------------------------------
 */
void ExecuteTclCommand(
	register Tcl_Interp *interp,	/* Current interpreter. */
	int tty,
	Boolean WaitFlag )
{
	extern TCL_COMMAND *TclCommandList;
	extern Semaphore ServerSemaphore;
	TCL_COMMAND *TclCommand;
	int code;
	PRIORITY_TYPE CurPri;

	CurPri = ProcSetPri( High );
	while ( NumTclPending > 0  ){
		while (!QEmpty( TclCommandList )){
			SemWait( &TclSemaphore );
			DeQueue( TclCommandList, TclCommand );
			SemSignal( &TclSemaphore );
			if ( TclCommand == NULL )
				return;
			if (( TclCommand->Command != NULL )&&( TclCommand->Command[0] != '\0' )){
				code = Tcl_RecordAndEval( interp, TclCommand->Command, 0 );
				if (code != TCL_OK) {
					write( 2, interp->result, strlen(interp->result)+1);
					write( 2, "\n", 1 );
				} else if (tty && (*interp->result != 0)) {
					write( 1, interp->result, strlen(interp->result)+1);
					write( 1, "\n", 1 );
				}
			}
			SemWait( &TclSemaphore );
			NumTclPending--;
			SemSignal( &TclSemaphore );
			vcp_free( TclCommand );
			if ( TclCommand->Command != NULL )
				vcp_free( TclCommand->Command );
			SemSignal( &ServerSemaphore );/* Allow server to run */
			SemWait( &ServerSemaphore );
		}
		if ( WaitFlag ){
			SemSignal( &ServerSemaphore );/* Allow server to run */
			ProcSleep(&TclProcDesc);
			SemWait( &ServerSemaphore );
		}
		else break;
	}
	SemSignal( &ServerSemaphore );/* Allow server to run */
	SemWait( &ServerSemaphore );
	ProcSetPri(CurPri);
}

/*}}}*/

/*{{{  Memory allocation and free routines*/
#include <node.h>
Semaphore AllocSemaphore = {1, 0, 0};
void *vcp_malloc( int size )
{
	void *ptr;
	SemWait( &AllocSemaphore );
	ptr = (void *)malloc( size );
	SemSignal( &AllocSemaphore );
	if ( ptr == NULL ) write( 2, "malloc failed\n", 14 );
	if (( ptr < _Node.process->heap )||( ptr > (char *)(_Node.process->heap) + _Node.process->heapsize ))
		fprintf( stderr, "malloc returned %x\n", ptr );
	return( ptr );
}

void vcp_free( void *ptr )
{
	SemWait( &AllocSemaphore );
	free(ptr);
	SemSignal( &AllocSemaphore );
}

void *vcp_realloc( void *ptr, int size )
{
	void *newptr;
	SemWait( &AllocSemaphore );
	newptr = (void *)realloc( ptr, size );
	SemSignal( &AllocSemaphore );
	if ( newptr == NULL ) write( 2, "realloc failed\n", 15 );
	if (( newptr < _Node.process->heap )||( newptr > (char *)(_Node.process->heap) + _Node.process->heapsize ))
		fprintf( stderr, "realloc returned %x\n", newptr );
	return( newptr );
}

/*}}}*/
