/*{{{  include files*/
#include <stdio.h>
#ifdef INMOSC
#include <channel.h>
#include <process.h>
#include <semaphor.h>
#else
#include <conc.h>
#endif
#include <string.h>
#include <stdlib.h>
#include "vcptype.h"
#include "tclVcpDefs.h"
#include "tclVcpType.h"
#include "queue.h"
#include "vcpmac.h"
#include "vcpext.h"
#include "tclVcpMac.h"
#include "tclVcpExt.h"
#include "debug.h"

Semaphore TclSemaphore = {1, 0, 0};
TCL_COMMAND *TclCommandList = NULL;
extern int NumTclPending;
/*}}}*/

/*{{{  QueueTclCommand*/
/*
 *------------------------------------------------------------------------
 *
 *	QueueTclCommand:
 *		Queues a TCL command,to be executed when TCL is idle
 *
 *	Returns nothing
 *
 *------------------------------------------------------------------------
 */
void QueueTclCommand( char *Command )
{
	QUEUE_TCL_COMMAND( Command );
}
/*}}}*/

/*{{{  VCP_ERROR*/

char *VCP_ErrorMessage[] = {
			"",
			"Packet Too Long",
			"Packet Too Short",
			"Unexpected Packet, ACK not yet sent",
			"Unexpected Ack",
			"Out Of Memory",
			"Internal Error",
			"VCP Not Initialized",
			"Illegal Link Number",
			"Receiving everything from link",
			"VCP timed out",
			"VCP/server already initialized",
			"Cannot open file",
			"Error in vcpxfer command format"
			};

/*
 *------------------------------------------------------------------------
 *
 *	VCP_ERROR:
 *		Print an error message for the specified Error number.
 *
 *	Return value :
 *		Error Number
 *
 *------------------------------------------------------------------------
 */
ERROR_TYPE VCP_ERROR( ERROR_TYPE ErrNumber )
{
	char *buf = malloc(100);
	sprintf( buf, "puts stderr \"VCP ERROR %d: %s\n\"", ErrNumber,  VCP_ErrorMessage[0-ErrNumber]);
	QueueTclCommand( buf );
	return( ErrNumber );
}
/*}}}*/

/*{{{  StrToNum*/
/*
 *------------------------------------------------------------------------
 *
 *	StrToNum:
 *		Computes the value of a number represented in a specified 
 *		radix (<=16) from the given string. Checks for validity
 *		are not made.
 *
 *	Return value :
 *		Value of the string
 *
 *------------------------------------------------------------------------
 */
unsigned long StrToNum( char *string, int radix )
{
	unsigned int val, temp;
	char c;
	val = 0;
	while (( !END_OF_ARG(string))&&( *string != ADDRESS_SEPARATOR)){
		c = *string; string++;
		if (( c >= '0' )&&( c <= '9' ))
			temp = c - '0';
		else if (( c >= 'a' )&&( c <= 'f' ))
			temp = c - 'a' + 10;
		else if (( c >= 'A' )&&( c <= 'F' ))
			temp = c - 'A' + 10;
		else 
			break;
		if ( temp >= radix )
			break;
		val = val * radix + temp;
	}
	return(val);
}
/*}}}*/

/*{{{  VCP_BlockedReceive*/

/*
 *------------------------------------------------------------------------
 *
 *	VCP_BlockedReceive :
 *		Receive a message and store it in the specified file or
 *	in a buffer.
 *	
 *	Parameters :
 *		ReceiveAddress : 
 *			A virtual channel address of the form <Link>.<Address>
 *			Address is assumed to be 16 bit if it is >255 or if
 *			starts with '0'. Link should be specified as L0-L3
 *		SendAddress :
 *			A virtual channel destination of the form
 *			(<Link>.)(<Header>.)*<Header>
 *			Link number is optional.
 *		FileOrBuffer :
 *			FileFlag = True implies receive into a file. In this
 *			case, BufferPtr specifies the filename.
 *			If FileFlag = False, the message is received into a
 *			buffer and the address of the buffer is returned in
 *			BufferPtr. Message is returned in a string of the
 *			form (<Hexbyte> )*
 *		TimeOut :
 *			Maximum number of milliseconds to wait for the message
 *		TimeOutCommand :
 *			TCL command to be executed on timeout
 *
 *	Return value :
 *		VCP_OK if successful, an Error number otherwise.
 *
 *------------------------------------------------------------------------
 */
int VCP_BlockedReceive( char *ReceiveAddress, char *SendAddress, 
	FILE_OR_BUFFER *FileOrBuffer, int TimeOut, char *TimeOutCommand  )
{
	int	LinkNumber;
	VIRTUAL_CHAN ReceiveHeader;
	HEADER	SendHeader;
	FILE	*fp;
	Boolean	EOMFlag;
	int	bufsize;
	int	ErrorFlag;
	char	*tempBuffer;
	BUFFER_LIST *BufferList = NULL;
	BUFFER_LIST *tempbuf;
	int	MessageLength = 0;

	tempBuffer = ( char *)vcp_malloc( MAX_BUF_SIZE );
	if ( tempBuffer == NULL )
		return( OutOfMemory );
	GET_LINK_NUMBER( ReceiveAddress, LinkNumber );
	FORM_VIRTUAL_CHAN_ADDR( ReceiveAddress, ReceiveHeader );
	FORM_HEADER( SendAddress, SendHeader );
	InitQ( BufferList );
	if ( FileOrBuffer->FileFlag ) {
		fp = fopen( FileOrBuffer->BufferPtr, "w" );
		if ( fp == NULL ){
			vcp_free( tempBuffer );
			return( CannotOpenFile );
		}
	}
	do {
		bufsize = MAX_BUF_SIZE;
		ErrorFlag = (int)VCP_ReceiveMessage( LinkNumber, ReceiveHeader, &SendHeader, tempBuffer, &bufsize, &EOMFlag, TimeOut*TICKS_PER_MILLISEC );
		if ( ErrorFlag == VCP_TimeOut ){
			SemWait( &TclSemaphore );
			NumTclPending++;
			SemSignal( &TclSemaphore );
			QUEUE_TCL_COMMAND( TimeOutCommand );
		}
		if ( ErrorFlag != VCP_OK )
			break;
		if (( bufsize != MAX_BUF_SIZE )&&( EOMFlag != True ))
			ErrorFlag = InternalError;
		if ( FileOrBuffer->FileFlag ) 
			fwrite( tempBuffer, bufsize, sizeof(char), fp );
		else {
			MEM_ALLOCATE( tempbuf, BUFFER_LIST, ErrorFlag, break );
			tempbuf->BufferSize = bufsize;
			tempbuf->BufferPtr = tempBuffer;
			MessageLength += bufsize;
			EnQueue( BufferList, tempbuf );
			if ( EOMFlag == True )
				break;
			tempBuffer = ( char *)vcp_malloc( MAX_BUF_SIZE );
			if ( tempBuffer == NULL ) {
				ErrorFlag =  OutOfMemory ;
				break;
			}
		}
	} while ( EOMFlag != True );
	if ( FileOrBuffer->FileFlag ) {
		fclose(fp);
		vcp_free( tempBuffer );
	}
	else {
		if ( ErrorFlag != VCP_OK )
			vcp_free( tempBuffer );
		if ( MessageLength == 0 ){
			tempBuffer = vcp_malloc( sizeof(char));
			tempBuffer[0] = '\0';
		}
		else
			tempBuffer  = (char *)vcp_malloc( MessageLength * LENGTH_PER_BYTE+1 );
		FileOrBuffer->BufferPtr = tempBuffer;
		while ( !QEmpty( BufferList )) {
			DeQueue( BufferList, tempbuf );
			MESSAGE_COPY( tempBuffer, tempbuf->BufferPtr, tempbuf->BufferSize );
			tempBuffer += tempbuf->BufferSize * LENGTH_PER_BYTE;
			vcp_free( tempbuf->BufferPtr );
			vcp_free( tempbuf );
		}
	}
	free( SendHeader.HeaderPtr );
	return( ErrorFlag );
}

/*}}}*/

/*{{{  VCP_ReceiveThread*/
/*
 *------------------------------------------------------------------------
 *
 *	VCP_ReceiveThread :
 *		Invokes VCP_BlockedReceive to receive a message into a file
 *	or buffer. When the receive is complete, this thread executes the
 *	TCL command specified in the parameter Command, and then terminates.
 *
 *	Returns nothing.
 *
 *------------------------------------------------------------------------
 */
static void VCP_ReceiveThread( int FIXME, char *ReceiveAddress, char *SendAddress, FILE_OR_BUFFER *FileOrBuffer, char *Command, char *RecvVar, int TimeOut, char *TimeOutCommand  )
{
	char *SetRecvVar;

	if ( VCP_BlockedReceive( ReceiveAddress, SendAddress, FileOrBuffer, TimeOut, TimeOutCommand ) != VCP_TimeOut ){
		/* Set TCL variable if required */
		if ( !FileOrBuffer->FileFlag ){
			int len;
			len = strlen( RecvVar )+ strlen( FileOrBuffer->BufferPtr ) +9;
			SetRecvVar = vcp_malloc(len *sizeof(char));
			sprintf( SetRecvVar, "set %s \"%s\"", RecvVar, FileOrBuffer->BufferPtr );
			QUEUE_TCL_COMMAND( SetRecvVar );
		}
		QUEUE_TCL_COMMAND( Command );
	}
	else {
		SemWait( &TclSemaphore );
		if ( FileOrBuffer->FileFlag )
			NumTclPending--;
		else
			NumTclPending -= 2;
		SemSignal( &TclSemaphore );
	}
	vcp_free( ReceiveAddress );
	vcp_free( SendAddress );
	if ( FileOrBuffer->FileFlag )
		vcp_free( FileOrBuffer->BufferPtr );
	vcp_free( FileOrBuffer );
	if ( TimeOutCommand != NULL )vcp_free( TimeOutCommand );
}
/*}}}*/

/*{{{  ReceiveAllThread*/

/*
 *------------------------------------------------------------------------
 *
 *	VCP_ReceiveAllThread :
 *		Receives any messages that have been buffered in MessageList,
 *	and sleeps after storing its descriptor in ReceiveWait. This thread
 *	is woken up by the VCP receiver thread whenever any message arrives.
 *	The messages are removed from MessageList and printed to the file fp.
 *
 *	Returns nothing.
 *
 *------------------------------------------------------------------------
 */
static void ReceiveAllThread( int FIXME, FILE *fp, Boolean *VCP_Running, MESSAGE_LIST **MessageList, WAIT_STRUCT *ReceiveWait )
{
	PRIORITY_TYPE CurrentPriority;
	MESSAGE_LIST *tempMessage;
	int MessageId = 0;

	while (*VCP_Running == True) {
		while ( !QEmpty(( *MessageList ))) {
			CurrentPriority = ProcSetPri( High );
			DeQueue( (*MessageList), tempMessage );
			ProcSetPri( CurrentPriority );
			fprintf( fp, "\nMESSAGE <%d>\n", MessageId++ );
			WRITE_MESSAGE_TO_FILE( tempMessage->MessagePtr, 
				(PACKET_LENGTH(tempMessage->MessagePtr)+1), fp );
			vcp_free( tempMessage->MessagePtr );
			vcp_free( tempMessage );
		}
		ReceiveWait->SleepFlag = True;
		ProcSleep( &(ReceiveWait->ProcessDescriptor) );
	}
	fclose( fp );

}
/*}}}*/

/*{{{  VCP_ReceiveAll*/

/*
 *------------------------------------------------------------------------
 *
 *	VCP_ReceiveAll :
 *		Informs VCP that all messages arriving on the specified
 *	link should be received. Creates a thread to copy the messages
 *	to a file.
 *
 *	Return value :
 *		VCP_OK if successful, error number otherwise.
 *
 *------------------------------------------------------------------------
 */
int VCP_ReceiveAll( int LinkNumber, char *FileName )
{
	MESSAGE_LIST **tempMessage;
	Process *pid;
	WAIT_STRUCT *Wait;
	Boolean *RunFlag;
	FILE *fp;

	tempMessage = (MESSAGE_LIST **)VCP_ReceiveEverything( LinkNumber, &Wait, &RunFlag );
	if (( Wait == NULL )||( RunFlag == NULL ))
		return( VCP_ReceiveAllSet );
	fp = fopen( FileName, "w" );
	if ( fp == NULL )
		return( CannotOpenFile );
	pid = ProcAlloc( ReceiveAllThread, DEFAULT_STACKSIZE, 4, fp, RunFlag, tempMessage, Wait );
	if ( pid == NULL )
		return( OutOfMemory );
	ProcRunHigh( pid );
	return( VCP_OK );
}
/*}}}*/

/*{{{  VCP_UnblockedReceive*/

/*
 *------------------------------------------------------------------------
 *
 *	VCP_UnBlockedReceive :
 *		Saves all the parameters and creates a thread to receive
 *	a message
 *
 *	Returns nothing.
 *
 *------------------------------------------------------------------------
 */
int VCP_UnBlockedReceive( char *ReceiveAddress, char *SendAddress, 
	FILE_OR_BUFFER *FileOrBuffer, char *Command, char *RecvVar, int TimeOut, char *TimeOutCommand  )
{
	Process *pid;
	char *tempReceiveAddress, *tempSendAddress, *tempCommand;
	FILE_OR_BUFFER *tempFileOrBuffer;
	char *tempFileName;
	char *tempRecvVar;
	char *tempTimeOutCommand;

	tempCommand = (char *)vcp_malloc( strlen(Command)+1 );
	tempReceiveAddress = ( char *)vcp_malloc( strlen(ReceiveAddress)+1 );
	tempSendAddress = ( char *)vcp_malloc( strlen(SendAddress)+1 );
	tempFileOrBuffer = MEM_ALLOC( FILE_OR_BUFFER );
	tempRecvVar = (char *)vcp_malloc( strlen(RecvVar)+1 );
	if (( !tempCommand )||( !tempReceiveAddress )||( !tempSendAddress )||
	    ( !tempFileOrBuffer ))
		return( OutOfMemory );
	strcpy( tempCommand, Command );
	strcpy( tempReceiveAddress, ReceiveAddress );
	strcpy( tempSendAddress, SendAddress );
	strcpy( tempRecvVar, RecvVar );
	*tempFileOrBuffer = *FileOrBuffer;
	if ( TimeOutCommand != NULL ){
		tempTimeOutCommand = (char *)vcp_malloc( strlen(TimeOutCommand)+1 );
		if ( !tempTimeOutCommand )
			return( OutOfMemory );
		strcpy( tempTimeOutCommand, TimeOutCommand );
	}
	else
		tempTimeOutCommand = NULL;
	if ( FileOrBuffer->FileFlag ) {
		tempFileName = (char *)vcp_malloc( strlen(FileOrBuffer->BufferPtr)+1 );
		strcpy( tempFileName, FileOrBuffer->BufferPtr );
		tempFileOrBuffer->BufferPtr = tempFileName;
	}
	pid = ProcAlloc( VCP_ReceiveThread, DEFAULT_STACKSIZE, 
		7, tempReceiveAddress, tempSendAddress, tempFileOrBuffer, 
		tempCommand, tempRecvVar, TimeOut, tempTimeOutCommand );
	if ( pid == NULL ) return( OutOfMemory );
	ProcRunHigh( pid );
	return( VCP_OK );

}
/*}}}*/

/*{{{  VCP_BlockedSend*/

/*
 *------------------------------------------------------------------------
 *
 *	VCP_BlockedSend :
 *		Send a message from a file or a specified buffer. The call
 *	blocks until the complete message is sent and the acknowledgement
 *	received.
 *	
 *	Parameters :
 *		SendAddress :
 *			A virtual channel destination of the form
 *			Header is assumed to be 16 bit if it is >255 or if
 *			starts with '0'. Link should be specified as L0-L3
 *			(<Link>.)(<Header>.)*<Header>
 *		ReceiveAddress : 
 *			A virtual channel address of the form <Link>.<Header>
 *			Link number is optional.
 *		FileOrBuffer :
 *			FileFlag = True implies send from a file. In this
 *			case, BufferPtr specifies the filename.
 *			If FileFlag = False, the data in BufferPtr is sent.
 *			Data in buffer is of the form (<HexByte> )*
 *		TimeOut :
 *			Maximum number of milliseconds to wait for the message 
 *			to be sent
 *		TimeOutCommand :
 *			TCL command to be executed on timeout
 *
 *	Return value :
 *		VCP_OK if successful, an Error number otherwise.
 *
 *------------------------------------------------------------------------
 */
int VCP_BlockedSend( char *SendAddress, char *ReceiveAddress, FILE_OR_BUFFER *FileOrBuffer, int TimeOut, char *TimeOutCommand  )
{
	int	LinkNumber;
	VIRTUAL_CHAN ReceiveHeader;
	HEADER	SendHeader;
	FILE	*fp;
	Boolean	EOMFlag;
	int	bufsize;
	int	ErrorFlag;
	unsigned char	*tempBuffer;
	BUFFER_LIST *BufferList = NULL;
	BUFFER_LIST *tempbuf;
	int	MessageLength = 0;

	GET_LINK_NUMBER( SendAddress, LinkNumber );
	FORM_VIRTUAL_CHAN_ADDR( ReceiveAddress, ReceiveHeader );
	FORM_HEADER( SendAddress, SendHeader );
	if ( !FileOrBuffer->FileFlag ) {
		bufsize = strlen(FileOrBuffer->BufferPtr);
		MAKE_MESSAGE( tempBuffer, FileOrBuffer->BufferPtr, bufsize, MessageLength );
		ErrorFlag = VCP_SendMessage( LinkNumber, &SendHeader, ReceiveHeader,
			tempBuffer, MessageLength, True, TimeOut*TICKS_PER_MILLISEC );
		if ( ErrorFlag == VCP_TimeOut ){
			SemWait( &TclSemaphore );
			NumTclPending++;
			SemSignal( &TclSemaphore );
			QUEUE_TCL_COMMAND( TimeOutCommand );
		}
	}
	else {
		EOMFlag = False;
		tempBuffer = ( unsigned char *)vcp_malloc( MAX_BUF_SIZE );
		if ( tempBuffer == NULL )
			return( OutOfMemory );
		fp = fopen( FileOrBuffer->BufferPtr, "r" );
		if ( fp == NULL ){
			vcp_free( tempBuffer );
			return( CannotOpenFile );
		}
		do {
			bufsize = fread( tempBuffer, sizeof(char), MAX_BUF_SIZE, fp );
			if ( bufsize <= 0 )
				break;
			if ( feof(fp))
				EOMFlag = True;
			ErrorFlag = VCP_SendMessage( LinkNumber, &SendHeader, 
				ReceiveHeader, tempBuffer, bufsize, EOMFlag,
				TimeOut * TICKS_PER_MILLISEC );
			if ( ErrorFlag == VCP_TimeOut ){
				SemWait( &TclSemaphore );
				NumTclPending++;
				SemSignal( &TclSemaphore );
				QUEUE_TCL_COMMAND( TimeOutCommand );
			}
			if ( ErrorFlag != VCP_OK )
				break;
		} while ( EOMFlag != True );
		fclose(fp);
		vcp_free( tempBuffer );
	}
	vcp_free( SendHeader.HeaderPtr );
	return( ErrorFlag );
}
/*}}}*/

/*{{{  VCP_SendThread*/

/*
 *------------------------------------------------------------------------
 *
 *	VCP_SendThread :
 *		Invokes VCP_BlockedSend to send a message from a file
 *	or buffer. When the send is complete, this thread executes the
 *	TCL command specified in the parameter Command, and then terminates.
 *
 *	Returns nothing.
 *
 *------------------------------------------------------------------------
 */
static void VCP_SendThread( int FIXME, char *SendAddress, char *ReceiveAddress, FILE_OR_BUFFER *FileOrBuffer, char *Command, int TimeOut, char *TimeOutCommand  )
{
	TCL_COMMAND	*TclCommand;

	if ( VCP_BlockedSend( SendAddress, ReceiveAddress, FileOrBuffer, TimeOut, TimeOutCommand ) != VCP_TimeOut )
		QUEUE_TCL_COMMAND( Command )
	else {
		SemWait( &TclSemaphore );
		NumTclPending--;
		SemSignal( &TclSemaphore );
	}
	vcp_free( ReceiveAddress );
	vcp_free( SendAddress );
	if ( FileOrBuffer->FileFlag )
		vcp_free( FileOrBuffer->BufferPtr );
	vcp_free( FileOrBuffer );
	if ( TimeOutCommand != NULL )
		vcp_free( TimeOutCommand );
}
/*}}}*/

/*{{{  VCP_UnBlockedSend*/
/*
 *------------------------------------------------------------------------
 *
 *	VCP_UnBlockedSend :
 *		Saves all the parameters and creates a thread to send
 *	a message
 *
 *	Returns nothing.
 *
 *------------------------------------------------------------------------
 */
int VCP_UnBlockedSend( char *SendAddress, char *ReceiveAddress, 
	FILE_OR_BUFFER *FileOrBuffer, char *Command, int TimeOut, char *TimeOutCommand  )
{
	Process *pid;
	char *tempReceiveAddress, *tempSendAddress, *tempCommand;
	FILE_OR_BUFFER *tempFileOrBuffer;
	char *tempFileName;
	char *tempTimeOutCommand;

	tempCommand = (char *)vcp_malloc( strlen(Command)+1 );
	tempReceiveAddress = ( char *)vcp_malloc( strlen(ReceiveAddress)+1 );
	tempSendAddress = ( char *)vcp_malloc( strlen(SendAddress)+1 );
	tempFileOrBuffer = MEM_ALLOC( FILE_OR_BUFFER );
	if (( !tempCommand )||( !tempReceiveAddress )||( !tempSendAddress )||
	    ( !tempFileOrBuffer )){
		VCP_ERROR( OutOfMemory );
		return( OutOfMemory );
	}
	strcpy( tempCommand, Command );
	strcpy( tempReceiveAddress, ReceiveAddress );
	strcpy( tempSendAddress, SendAddress );
	*tempFileOrBuffer = *FileOrBuffer;
	if ( TimeOutCommand != NULL ){
		tempTimeOutCommand = (char *)vcp_malloc( strlen(TimeOutCommand)+1 );
		if ( !tempTimeOutCommand )
			return( OutOfMemory );
		strcpy( tempTimeOutCommand, TimeOutCommand );
	}
	else
		tempTimeOutCommand = NULL;
	if ( FileOrBuffer->FileFlag ) {
		tempFileName = (char *)vcp_malloc( strlen(FileOrBuffer->BufferPtr)+1 );
		strcpy( tempFileName, FileOrBuffer->BufferPtr );
		tempFileOrBuffer->BufferPtr = tempFileName;
	}
	pid = ProcAlloc( VCP_SendThread, DEFAULT_STACKSIZE, 
		6, tempSendAddress, tempReceiveAddress, tempFileOrBuffer, 
		tempCommand, TimeOut, tempTimeOutCommand );
	if ( pid == NULL ) return( OutOfMemory );
	ProcRunHigh( pid );
		
	return( VCP_OK );
}

/*}}}*/


/*{{{  VCP_BlockedXfer*/

/*
 *------------------------------------------------------------------------
 *
 *	VCP_BlockedXfer :
 *		Transfer messages to a virtual channel address, and
 *		receive the reply messages.
 *	
 *	Parameters :
 *		SendChan :
 *			A virtual channel destination of the form
 *			Header is assumed to be 16 bit if it is >255 or if
 *			starts with '0'. Link should be specified as L0-L3
 *			(<Link>.)(<Header>.)*<Header>
 *		SendAckChan : 
 *			A virtual channel address of the form <Link>.<Header>
 *			Link number is optional.
 *		ReceiveChan :
 *			A virtual channel address of the form <Link>.<Header>
 *			Link number is optional.
 *			A virtual channel destination of the form
 *			Header is assumed to be 16 bit if it is >255 or if
 *			starts with '0'. Link should be specified as L0-L3
 *			(<Link>.)(<Header>.)*<Header>
 *		ReceiveAckChan : 
 *			A virtual channel destination of the form
 *			Header is assumed to be 16 bit if it is >255 or if
 *			starts with '0'. Link should be specified as L0-L3
 *			(<Link>.)(<Header>.)*<Header>
 *			Link number is optional.
 *		Buffer :
 *			A list of messages to be sent to SendAckChan.
 *			Reply from ReceiveChan is received and ignored.
 *			Each message is of the form <Data>.<Data>. .. r
 *		TimeOut :
 *			Maximum number of milliseconds to wait for the message 
 *			to be sent
 *		TimeOutCommand :
 *			TCL command to be executed on timeout
 *
 *	Return value :
 *		VCP_OK if successful, an Error number otherwise.
 *
 *------------------------------------------------------------------------
 */
#define MESSAGE_BUF_LEN	256
char StaticBuffer[MESSAGE_BUF_LEN];
int GetNextMessage( char **source, char **destination,  int *MessageLen, Boolean *over )
{
	ERROR_TYPE ErrorFlag;
	int len, di;
	unsigned long val;
	unsigned char *valptr;
	int digval, radix, vallen, i;
	Boolean MessageEnd;
	char *dest, *sptr, ch;

	sptr = *source;
	if (*sptr == '\0'){ *MessageLen = 0; *over = True;  return( VCP_OK );}
	len = 0;
	ErrorFlag = VCP_OK;
	MessageEnd = False;
	while (( *sptr != '\0' )&&( ErrorFlag == VCP_OK )&& (!MessageEnd)) {
		ch = *sptr;
		while (( ch == ' ' )||( ch == '\n' )||( ch == '\t' )){
			sptr++; ch = *sptr;
		}
		switch ( ch ) {
		case DATA_SEPARATOR:
			len +=  (( *(sptr-1) == LONG_WORD_SPEC )?4:
				(*(sptr-1) == SHORT_WORD_SPEC )?2:1);
			sptr++;
			break;
		case MESSAGE_SEPARATOR:
			MessageEnd = True;
			sptr++;
			if ( *sptr == DATA_SEPARATOR ) sptr++;
			while (( *sptr == ' ' )||( *sptr == '\n' )||( *sptr == '\t' )){
				sptr++; 
			}
			if ( *sptr == '\0' )*over = True;
			break;
		default :
			if ( !IS_HEXVAL( ch )&&(  ch != 'x' )&&( !WORD_LEN_SPEC(ch)))
				ErrorFlag = XferFormatError;
			sptr++;
			break;
		}
	}
	if (( ErrorFlag != VCP_OK )||( len == 0 )){
		*MessageLen = 0; *over = True;
		return( ErrorFlag );
	}
	if ( len < MESSAGE_BUF_LEN ){
		*destination = dest = StaticBuffer;
	}
	else 
		*destination = dest = vcp_malloc( len * sizeof(char));
	if ( dest == NULL ) {
		*MessageLen = 0; *over = True;
		return( OutOfMemory );
	}
	di = 0;
	sptr = *source;
	while (( di < len )&&( ErrorFlag == VCP_OK )){
		ch = *sptr;
		while (( ch == ' ' )||( ch == '\n' )||( ch == '\t' )){
			sptr++; ch = *sptr; /* ignore white space */
		}
		if ( ch == '\0' ){
			ErrorFlag = XferFormatError; break;
		}
		if ( ch == MESSAGE_SEPARATOR ){
			sptr++; 
			if ( *sptr == DATA_SEPARATOR )sptr++;
			break;
		}
		if (( ch == '0' )&&( *(sptr+1) == 'x' )){
			radix = 16; sptr += 2; ch = *sptr;
		}
		else if ( ch == '0' ){
			radix = 8; sptr++; ch = *sptr;
		}
		else
			radix = 10;
		val = 0; valptr = (unsigned char *)&val;
		vallen = 1;
		while (( ch != DATA_SEPARATOR )&&( ch != '\0' )){
			if ( ch == LONG_WORD_SPEC ){
				vallen = 4; sptr++; 
				if ( *sptr != DATA_SEPARATOR )
					ErrorFlag = XferFormatError;
				break;
			}
			else if ( ch == SHORT_WORD_SPEC ){
				vallen = 2; sptr++; 
				if ( *sptr != DATA_SEPARATOR )
					ErrorFlag = XferFormatError; 
				break;
			}
			digval = HEXVAL(ch);
			if ( digval > radix ){
				ErrorFlag = XferFormatError; break;
			}
			else
				val = val * radix + digval;
			sptr++; ch = *sptr;
		}
		for ( i = 0; i < vallen; i++ ){
			dest[di++] = *valptr++;
		}
		if ( *sptr == DATA_SEPARATOR )sptr++;
		else { *over = True; break; }
	}
	if ( *sptr = MESSAGE_SEPARATOR )sptr++;
	if ( *sptr == DATA_SEPARATOR ) sptr++; 
	while (( *sptr == ' ' )||( *sptr == '\n' )||( *sptr == '\t' )){
		sptr++; 
	}
	*MessageLen = di;
	*source = sptr;
	return( VCP_OK );
}

int VCP_BlockedXfer( 
	char *SendChan, 
	char *SendAckChan, 
	char *ReceiveChan, 
	char *ReceiveAckChan, 
	char *Buffer, 
	int TimeOut, 
	char *TimeOutCommand  )
{
	int	LinkNumber;
	VIRTUAL_CHAN ReceiveHeader, SendAckHeader;
	HEADER	SendHeader, ReceiveAckHeader;
	Boolean	EOMFlag;
	int	bufsize;
	int	ErrorFlag;
	unsigned char	*MsgBuffer;
	int	MsgLength = 0;
	Boolean MsgComplete;
	int temp;

	GET_LINK_NUMBER( SendChan, LinkNumber );
	FORM_VIRTUAL_CHAN_ADDR( ReceiveChan, ReceiveHeader );
	FORM_HEADER( SendChan, SendHeader );
	FORM_VIRTUAL_CHAN_ADDR( SendAckChan, SendAckHeader );
	FORM_HEADER( ReceiveAckChan, ReceiveAckHeader );
	MsgComplete = False;
	while ( !MsgComplete ){
		ErrorFlag = GetNextMessage( &Buffer, &MsgBuffer, &MsgLength, &MsgComplete );
		if ( MsgComplete && ( MsgLength <= 0 )) break;
		if ( ErrorFlag != VCP_OK ){ VCP_ERROR(ErrorFlag);break; }
		
		ErrorFlag = VCP_XferMessage( LinkNumber, &SendHeader, SendAckHeader,
			ReceiveHeader, &ReceiveAckHeader, 
			MsgBuffer, MsgLength, TimeOut*TICKS_PER_MILLISEC );
		if ( MsgBuffer != StaticBuffer ) vcp_free( MsgBuffer );
		if ( ErrorFlag == VCP_TimeOut ){
			ADD_TCL_COMMAND( TimeOutCommand );break;
		}
		if ( ErrorFlag != VCP_OK )break;
		/* Update Timeout */
	}
	vcp_free( SendHeader.HeaderPtr );
	vcp_free( ReceiveAckHeader.HeaderPtr );
	return( ErrorFlag );
}
/*}}}*/

/*{{{  InspectReceiveQ*/
/*
 *------------------------------------------------------------------------
 *
 *	InspectReceiveQ :
 *		Get headers of pending messages
 *
 *	Return value:
 *		List of headers of pending messages
 *
 *------------------------------------------------------------------------
 */
char *InspectReceiveQ( int LinkNumber )
{
	return(VCP_InspectQ( LinkNumber ));
}
/*}}}*/

/*{{{  vcpinit*/
/*
 *------------------------------------------------------------------------
 *
 *	vcpinit :
 *		Initialize VCP on a link
 *
 *	Returns nothing.
 *
 *------------------------------------------------------------------------
 */
void vcpinit( int LinkNumber )
{
	VCP_Init( LinkNumber );
}
/*}}}*/

/*{{{  vcpquit*/
/*
 *------------------------------------------------------------------------
 *
 *	vcpquit :
 *		Terminate VCP on a link
 *
 *	Returns nothing.
 *
 *------------------------------------------------------------------------
 */
void vcpquit( int LinkNumber )
{
	VCP_Quit( LinkNumber );
}
/*}}}*/

